@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 use base qw( IO::Async::Protocol::Stream );
 
@@ -78,6 +78,29 @@ behaviours and limits of its ability to communicate with Cassandra.
 
 =cut
 
+=head1 EVENTS
+
+=head2 on_event $name, @args
+
+A registered event occurred. C<@args> will depend on the event name. Each
+is also available as its own event, with the name in lowercase. If the event
+is not one of the types recognised below, C<@args> will contain the actual
+L<Protocol::CassandraCQL::Frame> object.
+
+=head2 on_topology_change $type, $node
+
+The cluster topology has changed. C<$node> is a packed socket address.
+
+=head2 on_status_change $status, $node
+
+The node's status has changed. C<$node> is a packed socket address.
+
+=head2 on_schema_change $type, $keyspace, $table
+
+A keyspace or table schema has changed.
+
+=cut
+
 =head1 PARAMETERS
 
 The following named parameters may be passed to C<new> or C<configure>:
@@ -92,6 +115,12 @@ The hostname of the Cassandra node to connect to
 
 Optional. The service name or port number to connect to.
 
+=item username => STRING
+
+=item password => STRING
+
+Optional. Authentication details to use for C<PasswordAuthenticator>.
+
 =item keyspace => STRING
 
 Optional. If set, a C<USE keyspace> query will be issued as part of the
@@ -120,7 +149,8 @@ sub configure
    my $self = shift;
    my %params = @_;
 
-   foreach (qw( host service keyspace default_consistency )) {
+   foreach (qw( host service username password keyspace default_consistency 
+                on_event on_topology_change on_status_change on_schema_change )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
    }
 
@@ -256,6 +286,8 @@ sub on_read
 
    # TODO: flags
    if( my $f = $self->{streams}[$streamid] ) {
+      undef $self->{streams}[$streamid];
+
       if( $opcode == OPCODE_ERROR ) {
          my $err     = $frame->unpack_int;
          my $message = $frame->unpack_string;
@@ -265,13 +297,19 @@ sub on_read
          $f->done( $opcode, $frame );
       }
 
-      undef $self->{streams}[$streamid];
-
       if( my $next = shift @{ $self->{pending} } ) {
          my ( $opcode, $frame, $f ) = @$next;
          $self->_send( $opcode, $streamid, $frame, $f );
       }
    }
+   elsif( $streamid == 0 and $opcode == OPCODE_ERROR ) {
+      my $err     = $frame->unpack_int;
+      my $message = $frame->unpack_string;
+      $self->fail_all_and_close( "OPCODE_ERROR: $message\n", $err, $frame );
+   }
+   elsif( $streamid == 0xff and $opcode == OPCODE_EVENT ) {
+      $self->_event( $frame );
+   }
    else {
       print STDERR "Received a message opcode=$opcode for unknown stream $streamid\n";
    }
@@ -279,6 +317,35 @@ sub on_read
    return 1;
 }
 
+sub _event
+{
+   my $self = shift;
+   my ( $frame ) = @_;
+
+   my $name = $frame->unpack_string;
+
+   my @args;
+   if( $name eq "TOPOLOGY_CHANGE" ) {
+      push @args, $frame->unpack_string; # type
+      push @args, $frame->unpack_inet;   # node
+   }
+   elsif( $name eq "STATUS_CHANGE" ) {
+      push @args, $frame->unpack_string; # status
+      push @args, $frame->unpack_inet;   # node
+   }
+   elsif( $name eq "SCHEMA_CHANGE" ) {
+      push @args, $frame->unpack_string; # type
+      push @args, $frame->unpack_string; # keyspace
+      push @args, $frame->unpack_string; # table
+   }
+   else {
+       push @args, $frame;
+   }
+
+   $self->maybe_invoke_event( "on_".lc($name), @args )
+      or $self->maybe_invoke_event( on_event => $name, @args );
+}
+
 sub on_closed
 {
    my $self = shift;
@@ -382,12 +449,47 @@ sub startup
       } )
    )->then( sub {
       my ( $op, $response ) = @_;
-      $op == OPCODE_READY or return $self->fail_all_and_close( "Expected OPCODE_READY" );
 
-      return Future->new->done;
+      if( $op == OPCODE_READY ) {
+         return Future->new->done;
+      }
+      elsif( $op == OPCODE_AUTHENTICATE ) {
+         return $self->_authenticate( $response->unpack_string );
+      }
+      else {
+         return $self->fail_all_and_close( "Expected OPCODE_READY or OPCODE_AUTHENTICATE" );
+      }
    });
 }
 
+sub _authenticate
+{
+   my $self = shift;
+   my ( $authenticator ) = @_;
+
+   if( $authenticator eq "org.apache.cassandra.auth.PasswordAuthenticator" ) {
+      foreach (qw( username password )) {
+         defined $self->{$_} or croak "Cannot authenticate by password without $_";
+      }
+
+      $self->send_message( OPCODE_CREDENTIALS,
+         Protocol::CassandraCQL::Frame->new->pack_string_map( {
+            username => $self->{username},
+            password => $self->{password},
+         }
+      )
+      )->then( sub {
+         my ( $op, $response ) = @_;
+         $op == OPCODE_READY or return $self->fail_all_and_close( "Expected OPCODE_READY" );
+
+         return Future->new->done;
+      });
+   }
+   else {
+      return $self->fail_all_and_close( "Unrecognised authenticator $authenticator" );
+   }
+}
+
 =head2 $f = $cass->options
 
 Requests the list of supported options from the server node. On success, the
@@ -537,6 +639,29 @@ sub execute
    });
 }
 
+=head2 $f = $cass->register( $events )
+
+Registers the connection's interest in receiving events of the types given in
+the ARRAY reference. Event names may be C<TOPOLOGY_CHANGE>, C<STATUS_CHANGE>
+or C<SCHEMA_CHANGE>. On success, the returned Future yields nothing.
+
+=cut
+
+sub register
+{
+   my $self = shift;
+   my ( $events ) = @_;
+
+   $self->send_message( OPCODE_REGISTER,
+      Protocol::CassandraCQL::Frame->new->pack_string_list( $events )
+   )->then( sub {
+      my ( $op, $response ) = @_;
+      $op == OPCODE_READY or Future->new->fail( "Expected OPCODE_READY" );
+
+      return Future->new->done;
+   });
+}
+
 =head1 CONVENIENT WRAPPERS
 
 The following wrapper methods all wrap the basic C<query> operation.
@@ -641,10 +766,6 @@ sub schema_columns
 
 =item *
 
-Handle OPCODE_AUTHENTICATE and OPCODE_REGISTER
-
-=item *
-
 Support frame compression
 
 =item *
